from typing import Union

from django.db import transaction, connection

from .models import Ticket
from . import NotInTransactionException


def acquire(semaphore_name: str) -> Union[Ticket, None]:
    """
    Acquire a ticket for a semaphore. Returns the ticket if successful, and None if there are none available.
    It is the callers responsibility to be inside a transaction context.
    """
    if not transaction.get_connection().in_atomic_block:
        raise NotInTransactionException('Semaphore tickets must be acquired inside a DB transaction block')
    return Ticket.objects.order_by('pk').select_for_update(skip_locked=True).filter(semaphore_name=semaphore_name).first()


@transaction.atomic
def make(semaphore_name: str, amt_tickets: int = 1):
    """
    Creates a new semaphore or adjusts the number of tickets of an existing one.
    Locks the table so that only one such adjustment can be done at a time.
    A downwards adjustment involves deletion of ticket rows, this blocks on rows
    representing currently issued tickets. Thus, the downward adjustment will take place
    once enough tickets have been returned to the semaphore.
    """
    connection.cursor().execute(f"LOCK TABLE {connection.ops.quote_name(Ticket._meta.db_table)} IN SHARE ROW EXCLUSIVE MODE")
    extants = Ticket.objects.filter(semaphore_name=semaphore_name).values_list('pk', flat=True)
    if len(extants) == amt_tickets:
        return  # Just right, nothing to do
    elif len(extants) < amt_tickets:
        Ticket.objects.bulk_create((Ticket(semaphore_name=semaphore_name) for i in range(amt_tickets - len(extants))))
    else:
        Ticket.objects.filter(pk__in=extants[amt_tickets:]).delete()


def destroy(semaphore_name: str):
    """
    Removes a semaphore.
    """
    make(semaphore_name, 0)


def list():
    """
    Returns a list of available semaphores and the number of tickets they each hold.
    """
    with connection.cursor() as cur:
        cur.execute(f"""
            SELECT
                semaphore_name,
                count(id)
            FROM
                {connection.ops.quote_name(Ticket._meta.db_table)}
            GROUP BY
                semaphore_name;
        """)
        return dict(cur)



from django.db import models


class Ticket(models.Model):
    semaphore_name = models.CharField(null=False, blank=False, max_length=255, help_text="The name of the semaphore. Create as many rows as you want with the same name; the semaphore by that name will hold that many tickets.")

    def __str__(self):
        return f'Ticket {self.semaphore_name}@{self.id}'



from django.contrib import admin

from .models import Ticket


@admin.register(Ticket)
class TicketAdmin(admin.ModelAdmin):
    list_filter = ('semaphore_name',)
    list_display = ('__str__', 'semaphore_name',)
    ordering = ('semaphore_name', 'pk')



# Generated by Django 3.2.13 on 2022-08-30 18:14

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Ticket',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('semaphore_name', models.CharField(help_text='The name of the semaphore. Create as many rows as you want with the same name; the semaphore by that name will hold that many tickets.', max_length=255)),
            ],
        ),
    ]



from subprocess import Popen, PIPE, DEVNULL
from sys import executable as this_python

from django.test.testcases import TransactionTestCase
from django.conf import settings

from .semaphore import make, list

TEST_SEMAPHORE = "Test"


class SemaphoreTest(TransactionTestCase):
    @classmethod
    def setUp(cls):
        # Create a semaphore with 2 tickets
        make(TEST_SEMAPHORE, 2)

    def test_list(self):
        self.assertEqual(list(), {TEST_SEMAPHORE: 2})

    def test_enlarge(self):
        make(TEST_SEMAPHORE, 3)
        self.assertEqual(list(), {TEST_SEMAPHORE: 3})

    def test_shrink(self):
        make(TEST_SEMAPHORE, 2)
        self.assertEqual(list(), {TEST_SEMAPHORE: 2})

    def test_acquire(self):
        """
        Launch three independent processes, each of which
        will attempt to acquire a ticket from the semaphore.
        We expect two to succeed and one to fail.
        """
        call_argv = [this_python, 'manage.py', 'dbsemaphore-test-acquire', '--dbname', settings.DATABASES['default']['NAME'], TEST_SEMAPHORE]
        launched = [Popen(call_argv, stdin=PIPE, stdout=PIPE, stderr=DEVNULL, text=True) for i in range(3)]
        outcomes = {int(p.stdout.read(1)) for p in launched}
        for p in launched:
            p.communicate(input="\n")
        self.assertEqual(outcomes, {0, 1, 2})



from time import sleep
from sys import stderr

from django.core.management.base import BaseCommand
from django.db import transaction
from django.conf import settings

from dbsemaphore.semaphore import acquire


class Command(BaseCommand):
    help = """Test a semaphore"""

    def add_arguments(self, parser):
        parser.add_argument('-n', '--dbname', help="Database name (useful for tests)")
        parser.add_argument('name', nargs=1, help="Name of semaphore")
        parser.add_argument('release', nargs='?', type=float, help="Release after X seconds")

    def handle(self, *args, **options):
        try:
            if options['dbname']:
                settings.DATABASES['default']['NAME'] = options['dbname']
            with transaction.atomic():
                ticket = acquire(options['name'][0])
                if ticket:
                    print(ticket.pk)
                else:
                    print(0)
                if options['release'] is not None:
                    sleep(options['release'])
                else:
                    print("Gently press Enter to release and exit…", file=stderr, end='')
                    input()
        except KeyboardInterrupt:
            exit("Interrupted")
